=encoding utf8
=pod

=head1 Description

PORCUS - Parsing de Oraciones Realizado por Computadora, Unificado y Serial
(Unified and Serialized Computer-based Sentence Parsing)

This software provides routines to execute POS taggers with one command. The 
purpose of this is to have in a single place any POS tagger that you use, in a 
simple and easy way. Also, PORCUS calculates the time that every tagger takes 
in its execution, so you can compare their perfomance.

For now, we only support English and Spanish languages and we provide you 
with six POS taggers to download separately: TreeTagger, UDPipe, spaCy, 
IXA-pipes, Stanford CoreNLP and DepPattern. However, you can install more in your 
own system in the ./taggers directory. See the modules' files to get examples on 
how to create your own.

We have an web demo to show PORCUS' performance: www.tecling.com/porcus
In this demo, we add SyntaxNet from Google, LSTM-parser, FreeLing and Connexor.
For more information, run perl /path/to/porcus.pl --help in the command line.

=head1 System Requirements

* UNIX-like OS system
  (We have tested PORCUS in Ubuntu, Arch Linux and Ubuntu in Windows Subsystem Linux)
  
* Perl 6

* Python >= 2.7

* Java 8 (>= 1.8.0)

=head1 Usage

Run perl porcus.pl --help in the PORCUS' directory.

=head1 License

PORCUS - Parsing de Oraciones Realizado por Computadora, Unificado y Serial
(Unified and Serialized Computer-based Sentence Parsing)

Nicolás Acosta - www.tecling.com/acosta & Rogelio Nazar - www.tecling.com/nazar

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

=head1 Contact

Thanks for using PORCUS!

Nicolás Acosta - nmac.1996@gmail.com
Universidad Nacional de Cuyo, Mendoza, Argentina

Rogelio Nazar - rogelio.nazar@gmail.com
Pontificia Universidad Católica de Valparaíso

http://www.tecling.com/

=head1 References

Connexor

  Pasi Tapanainen and Timo Järvinen. (1997). A non-projective dependency parser. In Proceedings of the fifth conference on Applied natural language processing (ANLC '97). Association for Computational Linguistics, Stroudsburg, PA, USA, 64-71.

DepPattern

  Gamallo P., González I. (2011) A Grammatical Formalism Based on Patterns of Part-of-Speech Tags, International Journal of Corpus Linguistics, 16(1), 45-71. ISNN:1384-6655
  
  Gamallo, P. 2015. Dependency Parsing with Compression Rules, The 14th International Conference on Parsing Technologies (IWPT-2015) p. 107-117, Bilbao. ISBN 978-1-941643-98-3
  
  Gamallo, P., González, I. 2012. DepPattern: A Multilingual Dependency Parser, Demo Session of the International Conference on Computational Processing of the Portuguese Language (PROPOR 2012), April 17-20, Coimbra, Portugal.
  
FreeLing

  Jordi Atserias and Bernardino Casas and Elisabet Comelles and Meritxell González and Lluís Padró and Muntsa Padró. FreeLing 1.3: Syntactic and semantic services in an open-source NLP library. Proceedings of the fifth international conference on Language Resources and Evaluation (LREC 2006), ELRA. Genoa, Italy. May, 2006.
  
  Jordi Carrera and Irene Castellón and Marina Lloberes and Lluís Padró and Nevena Tinkova. Dependency Grammars in FreeLing. Procesamiento del Lenguaje Natural n. 41, pg. 21--28. September, 2008.
  
  Marina Lloberes, Irene Castellón and Lluís Padró. Spanish FreeLing Dependency Grammar. Proceedings of the Seventh International Conference on Language Resources and Evaluation (LREC'10), La Valleta, Malta, 2010.
  
IXA-pipes

  Rodrigo Agerri, Josu Bermudez and German Rigau (2014): IXA pipes: Efficient and Ready to Use Multilingual NLP tools. In Proceedings of the 9th Language Resources and Evaluation Conference (LREC2014), 26-31 May, 2014, Reykjavik, Iceland.
  
LSTM-parser

  Chris Dyer and Miguel Ballesteros and Wang Ling and Austin Matthews and Noah A. Smith (2015). Transition-based Dependency Parsing with Stack Long Short-Term Memory. Proceedings of the 53rd Annual Meeting of the Association for Computational Linguistics and the 7th International Joint Conference on Natural Language Processing.
  
SpaCy

  Visit <spacy.io> to get documentation.
  
Stanford CoreNLP

  Christopher D. Manning (2011). Part-of-Speech Tagging from 97% to 100%: Is It Time for Some Linguistics? Proceedings of 12th International Conference, CICLing 2011, Tokyo, Japan, February 20-26, 2011: 171-189.
  
SyntaxNet

  Weiss, D., Alberti, C., Collins, M., & Petrov, S. (2015). Structured Training for Neural Network Transition-Based Parsing. Proceedings of the 53rd Annual Meeting of the Association for Computational Linguistics and the 7th International Joint Conference on Natural Language Processing.
  
TreeTagger

  Helmut Schmid (1995): Improvements in Part-of-Speech Tagging with an Application to German. Proceedings of the ACL SIGDAT-Workshop. Dublin, Ireland.
  
  Helmut Schmid (1994): Probabilistic Part-of-Speech Tagging Using Decision Trees. Proceedings of International Conference on New Methods in Language Processing, Manchester, UK.
  
UDPipe

  Milan Straka and Jana Straková (2017). Tokenizing, POS Tagging, Lemmatizing and Parsing UD 2.0 with UDPipe. In Proceedings of the CoNLL 2017 Shared Task: Multilingual Parsing from Raw Text to Universal Dependencies, Vancouver, Canada, August 2017.
  
  Straka Milan, Hajič Jan, Straková Jana (2016). UDPipe: Trainable Pipeline for Processing CoNLL-U Files Performing Tokenization, Morphological Analysis, POS Tagging and Parsing. In Proceedings of the Tenth International Conference on Language Resources and Evaluation (LREC 2016), Portorož, Slovenia, May 2016.
=cut

use strict;
use warnings;
use Cwd;
use Getopt::Long;
use Time::HiRes qw(gettimeofday);

our ($inputfile, $tagger, $language, $format, $id, $unlink, $verbose, $help, $perftime);

# Arguments of the script
GetOptions (
	"file=s" => \$inputfile,
	"tagger=s" => \$tagger,
	"lang=s" => \$language, 
	"format=s" => \$format,
	"id=s" => \$id,
	"unlink" => \$unlink,
	"verbose" => \$verbose,
	"help" => \$help,
);

# We define the main paths
our $home = getcwd;
our $modules = $home."/modules/";
our $tagpath = $home."/taggers/";
our $tmpath = '/tmp';

# We define table properties for html
our $tableprop = "<table class='table table-bordered table-striped table-sm'>";

# Dependencies of some taggers
our $python = '/usr/bin/python';
our $java = '/usr/bin/java';

# We set an id for document if it not defined by user
$id = int(rand(99)).int(rand(99)).int(rand(99)).int(rand(99)) if (!$id);

# A list of taggers that we have available
# Comment the tagger options that you have not yet
# installed adding a # at the start of the line.
our %taggers = (
	'treetagger' => 'TreeTagger',
	#'spacy' => 'spaCy',
	'deppattern' => 'DepPattern',
	'ixa' => 'IXA pipes',
	'stanford' => 'Stanford CoreNLP',
	'udpipe' => 'UDPipe',
);

# We create a list of available tagger to display in the help message
our $listagger = '';
foreach my $t (sort keys %taggers){
	$listagger .= "\n\t$taggers{$t} => [$t]";
}

# Usage of PORCUS
our $usage = "Usage:
perl porcus.pl --file /file/to/tag --tagger postagger --lang spanish [--format] [--id keyword] [--unlink] [--verbose] [--help]

--file
	Path of the file that you want to tag.
	Without this option, this program can't work.
--tagger
	Select one or more POS taggers to use. Separate multiple options with a comma without spaces.
	Example: --tagger treetagger,freeling,udpipe
	You can use the option 'all' to tag your text with all taggers available: --tagger all
	Without this option, this program can't work.
--lang
	Set language of the input file. For now, we only support [english] and [spanish].
	Without this option, this program can't work.
--format
	Select one output format file. Valid options: 'csv', 'html' or 'txt' (default: txt).
	csv: It exports plain text with separated values with tabs.
	html: It prints output of tagger in html text. Recommended for web usage.
	txt: It exports the default output format of each tagger.
--id
    Set an id for your document. It should not have spaces.
--unlink
	Remove tagged files after process finishes.
	(Only available for 'html' format option. This is unchecked by default).
--verbose
	Get detailed messages of execution of PORCUS.
--help
	Show this help message.

List of available POS taggers:$listagger
";

# We will check if all arguments are correct
$help = 1 if (!$inputfile || !$tagger || !$language);
if ($help){
	die $usage;
}

if ($tagger ne "all"){
	my @selectedtaggers;
	@selectedtaggers = split /,/, $tagger if ($tagger =~ /,/);
	if (@selectedtaggers){
		foreach my $st (@selectedtaggers){
			die "Sorry, [$st] isn't available. Try with one or more of these: $listagger\n" if (!$taggers{$st});
		}
	}else{
		die "Sorry, [$tagger] isn't available. Try with one or more of these: $listagger\n" if (!$taggers{$tagger});
	}
}

if (!-e $inputfile){
	die "Input file [$inputfile] doesn't exist.\n";
}

if ($language !~ /^(spanish|english)$/){
	die "Selected language [$language] isn't supported yet. Try [perl porcus.pl --help] for more information. Sorry!\n";
}

if ($format && $format !~ /^(txt|csv|html)$/){
	die "Selected format [$format] is invalid. Try [perl porcus.pl --help] for more information.\n";
}

# We will create output path
our $outpath = $home."/output/$id/";
if (!-d $outpath && $format ne "html"){
    mkdir $outpath;
}

# Set defaults
$format = 'txt' if (!$format);

# We load modules to work. They will execute selected taggers
foreach my $t (sort keys %taggers){
	warn "I cannot find module for [$t]\n" and next if (!-e $modules.$t.".pm");
	require ($modules.$t.".pm");
}

# Message for verbose mode when loading
sub loading {
	my $t = $_[0];
	my $com = $_[1];
	if ($format eq "html"){
		print "<hr><br><h5>Procesando texto con $t...</h5><br>\n";
	}else{
		warn "Procesando texto con $t...\n";
		warn "RUNNING: [$com]\n";
	}
}

# Calculating execution time of POS tagger
sub perftime {
	my $start = gettimeofday();
	system ($_[0]);
	my $end = gettimeofday();
	my $diff = $end - $start;
	$perftime = sprintf("%.5f", $diff);
	warn "$perftime seconds elapsed to complete tagging operation.\n" if ($verbose);
}
